/*
 * Copyright  Micromega Corporation 2002-2007. All rights reserved.
 */
package com.micromegacorp.math.v3_spi;
import  stamp.core.*;

/**
 * @file    Fpu.java
 * @target  Javelin Stamp
 *
 * This routine provides support for the uM-FPU V3.1 floating point
 * coprocessor connected using an SPI interface.
 *
 * @author Cam Thompson, Micromega Corporation
 * @version
 *    August 24, 2007
 *    - updated for uM-FPU V3.1
 *    March 19, 2006
 *    - original version for uM-FPU V3
 */

public class Fpu {

//-------------------- uM-FPU V3.1 opcode definitions --------------------------

  final public static int SELECTA   = 0x01;   // Select register A
  final public static int SELECTX   = 0x02;   // Select register X
  final public static int CLR       = 0x03;   // reg[nn] = 0
  final public static int CLRA      = 0x04;   // reg[A] = 0
  final public static int CLRX      = 0x05;   // reg[X] = 0, X = X + 1
  final public static int CLR0      = 0x06;   // reg[0] = 0
  final public static int COPY      = 0x07;   // reg[nn] = reg[mm]
  final public static int COPYA     = 0x08;   // reg[nn] = reg[A]
  final public static int COPYX     = 0x09;   // reg[nn] = reg[X], X = X + 1
  final public static int LOAD      = 0x0A;   // reg[0] = reg[nn]
  final public static int LOADA     = 0x0B;   // reg[0] = reg[A]
  final public static int LOADX     = 0x0C;   // reg[0] = reg[X], X = X + 1
  final public static int ALOADX    = 0x0D;   // reg[A] = reg[X], X = X + 1
  final public static int XSAVE     = 0x0E;   // reg[X] = reg[nn], X = X + 1
  final public static int XSAVEA    = 0x0F;   // reg[X] = reg[A], X = X + 1
  final public static int COPY0     = 0x10;   // reg[nn] = reg[0]
  final public static int COPYI     = 0x11;   // reg[nn] = long(unsigned bb)
  final public static int SWAP      = 0x12;   // Swap reg[nn] and reg[mm]
  final public static int SWAPA     = 0x13;   // Swap reg[A] and reg[nn]
  final public static int LEFT      = 0x14;   // Left parenthesis
  final public static int RIGHT     = 0x15;   // Right parenthesis
  final public static int FWRITE    = 0x16;   // Write 32-bit float to reg[nn]
  final public static int FWRITEA   = 0x17;   // Write 32-bit float to reg[A]
  final public static int FWRITEX   = 0x18;   // Write 32-bit float to reg[X], X = X + 1
  final public static int FWRITE0   = 0x19;   // Write 32-bit float to reg[0]
  final public static int FREAD     = 0x1A;   // Read 32-bit float from reg[nn]
  final public static int FREADA    = 0x1B;   // Read 32-bit float from reg[A]
  final public static int FREADX    = 0x1C;   // Read 32-bit float from reg[X], X = X + 1
  final public static int FREAD0    = 0x1D;   // Read 32-bit float from reg[0]
  final public static int ATOF      = 0x1E;   // Convert ASCII to float, store in reg[0]
  final public static int FTOA      = 0x1F;   // Convert float to ASCII
  final public static int FSET      = 0x20;   // reg[A] = reg[nn]
  final public static int FADD      = 0x21;   // reg[A] = reg[A] + reg[nn]
  final public static int FSUB      = 0x22;   // reg[A] = reg[A] - reg[nn]
  final public static int FSUBR     = 0x23;   // reg[A] = reg[nn] - reg[A]
  final public static int FMUL      = 0x24;   // reg[A] = reg[A] * reg[nn]
  final public static int FDIV      = 0x25;   // reg[A] = reg[A] / reg[nn]
  final public static int FDIVR     = 0x26;   // reg[A] = reg[nn] / reg[A]
  final public static int FPOW      = 0x27;   // reg[A] = reg[A] ** reg[nn]
  final public static int FCMP      = 0x28;   // Float compare reg[A] - reg[nn]
  final public static int FSET0     = 0x29;   // reg[A] = reg[0]
  final public static int FADD0     = 0x2A;   // reg[A] = reg[A] + reg[0]
  final public static int FSUB0     = 0x2B;   // reg[A] = reg[A] - reg[0]
  final public static int FSUBR0    = 0x2C;   // reg[A] = reg[0] - reg[A]
  final public static int FMUL0     = 0x2D;   // reg[A] = reg[A] * reg[0]
  final public static int FDIV0     = 0x2E;   // reg[A] = reg[A] / reg[0]
  final public static int FDIVR0    = 0x2F;   // reg[A] = reg[0] / reg[A]
  final public static int FPOW0     = 0x30;   // reg[A] = reg[A] ** reg[0]
  final public static int FCMP0     = 0x31;   // Float compare reg[A] - reg[0]
  final public static int FSETI     = 0x32;   // reg[A] = float(bb)
  final public static int FADDI     = 0x33;   // reg[A] = reg[A] + float(bb)
  final public static int FSUBI     = 0x34;   // reg[A] = reg[A] - float(bb)
  final public static int FSUBRI    = 0x35;   // reg[A] = float(bb) - reg[A]
  final public static int FMULI     = 0x36;   // reg[A] = reg[A] * float(bb)
  final public static int FDIVI     = 0x37;   // reg[A] = reg[A] / float(bb)
  final public static int FDIVRI    = 0x38;   // reg[A] = float(bb) / reg[A]
  final public static int FPOWI     = 0x39;   // reg[A] = reg[A] ** bb
  final public static int FCMPI     = 0x3A;   // Float compare reg[A] - float(bb)
  final public static int FSTATUS   = 0x3B;   // Float status of reg[nn]
  final public static int FSTATUSA  = 0x3C;   // Float status of reg[A]
  final public static int FCMP2     = 0x3D;   // Float compare reg[nn] - reg[mm]
  final public static int FNEG      = 0x3E;   // reg[A] = -reg[A]
  final public static int FABS      = 0x3F;   // reg[A] = | reg[A] |
  final public static int FINV      = 0x40;   // reg[A] = 1 / reg[A]
  final public static int SQRT      = 0x41;   // reg[A] = sqrt(reg[A])
  final public static int ROOT      = 0x42;   // reg[A] = root(reg[A], reg[nn])
  final public static int LOG       = 0x43;   // reg[A] = log(reg[A])
  final public static int LOG10     = 0x44;   // reg[A] = log10(reg[A])
  final public static int EXP       = 0x45;   // reg[A] = exp(reg[A])
  final public static int EXP10     = 0x46;   // reg[A] = exp10(reg[A])
  final public static int SIN       = 0x47;   // reg[A] = sin(reg[A])
  final public static int COS       = 0x48;   // reg[A] = cos(reg[A])
  final public static int TAN       = 0x49;   // reg[A] = tan(reg[A])
  final public static int ASIN      = 0x4A;   // reg[A] = asin(reg[A])
  final public static int ACOS      = 0x4B;   // reg[A] = acos(reg[A])
  final public static int ATAN      = 0x4C;   // reg[A] = atan(reg[A])
  final public static int ATAN2     = 0x4D;   // reg[A] = atan2(reg[A], reg[nn])
  final public static int DEGREES   = 0x4E;   // reg[A] = degrees(reg[A])
  final public static int RADIANS   = 0x4F;   // reg[A] = radians(reg[A])
  final public static int FMOD      = 0x50;   // reg[A] = reg[A] MOD reg[nn]
  final public static int FLOOR     = 0x51;   // reg[A] = floor(reg[A])
  final public static int CEIL      = 0x52;   // reg[A] = ceil(reg[A])
  final public static int ROUND     = 0x53;   // reg[A] = round(reg[A])
  final public static int FMIN      = 0x54;   // reg[A] = min(reg[A], reg[nn])
  final public static int FMAX      = 0x55;   // reg[A] = max(reg[A], reg[nn])
  final public static int FCNV      = 0x56;   // reg[A] = conversion(nn, reg[A])
  final public static int FMAC      = 0x57;   // reg[A] = reg[A] + (reg[nn] * reg[mm])
  final public static int FMSC      = 0x58;   // reg[A] = reg[A] - (reg[nn] * reg[mm])
  final public static int LOADBYTE  = 0x59;   // reg[0] = float(signed bb)
  final public static int LOADUBYTE = 0x5A;   // reg[0] = float(unsigned byte)
  final public static int LOADWORD  = 0x5B;   // reg[0] = float(signed word)
  final public static int LOADUWORD = 0x5C;   // reg[0] = float(unsigned word)
  final public static int LOADE     = 0x5D;   // reg[0] = 2.7182818
  final public static int LOADPI    = 0x5E;   // reg[0] = 3.1415927
  final public static int LOADCON   = 0x5F;   // reg[0] = float constant(nn)
  final public static int FLOAT     = 0x60;   // reg[A] = float(reg[A])
  final public static int FIX       = 0x61;   // reg[A] = fix(reg[A])
  final public static int FIXR      = 0x62;   // reg[A] = fix(round(reg[A]))
  final public static int FRAC      = 0x63;   // reg[A] = fraction(reg[A])
  final public static int FSPLIT    = 0x64;   // reg[A] = int(reg[A]), reg[0] = frac(reg[A])
  final public static int SELECTMA  = 0x65;   // Select matrix A
  final public static int SELECTMB  = 0x66;   // Select matrix B
  final public static int SELECTMC  = 0x67;   // Select matrix C
  final public static int LOADMA    = 0x68;   // reg[0] = matrix A[bb,bb]
  final public static int LOADMB    = 0x69;   // reg[0] = matrix B[bb, bb]
  final public static int LOADMC    = 0x6A;   // reg[0] = matrix C[bb, bb]
  final public static int SAVEMA    = 0x6B;   // matrix A[bb,bb] = reg[A]
  final public static int SAVEMB    = 0x6C;   // matrix B[bb,bb] = reg[A]
  final public static int SAVEMC    = 0x6D;   // matrix C[bb,bb] = reg[A]
  final public static int MOP       = 0x6E;   // Matrix operation
  final public static int FFT       = 0x6F;   // FFT operation
  final public static int WRBLK     = 0x70;   // write register block
  final public static int RDBLK     = 0x71;   // read register block
  final public static int LOADIND   = 0x7A;   // reg[0] = reg[reg[nn]]
  final public static int SAVEIND   = 0x7B;   // reg[reg[nn]] = reg[A]
  final public static int INDA      = 0x7C;   // Select A using reg[nn]
  final public static int INDX      = 0x7D;   // Select X using reg[nn]
  final public static int FCALL     = 0x7E;   // Call function in Flash memory
  final public static int EECALL    = 0x7F;   // Call function in EEPROM memory
  final public static int RET       = 0x80;   // Return from function
  final public static int BRA       = 0x81;   // Unconditional branch
  final public static int BRACC     = 0x82;   // Conditional branch
  final public static int JMP       = 0x83;   // Unconditional jump
  final public static int JMPCC     = 0x84;   // Conditional jump
  final public static int TABLE     = 0x85;   // Table lookup
  final public static int FTABLE    = 0x86;   // Floating point reverse table lookup
  final public static int LTABLE    = 0x87;   // Long integer reverse table lookup
  final public static int POLY      = 0x88;   // reg[A] = nth order polynomial
  final public static int GOTO      = 0x89;   // Computed goto
  final public static int RETCC     = 0x8A;   // Conditional return from function
  final public static int LWRITE    = 0x90;   // Write 32-bit long integer to reg[nn]
  final public static int LWRITEA   = 0x91;   // Write 32-bit long integer to reg[A]
  final public static int LWRITEX   = 0x92;   // Write 32-bit long integer to reg[X], X = X + 1
  final public static int LWRITE0   = 0x93;   // Write 32-bit long integer to reg[0]
  final public static int LREAD     = 0x94;   // Read 32-bit long integer from reg[nn]
  final public static int LREADA    = 0x95;   // Read 32-bit long integer from reg[A]
  final public static int LREADX    = 0x96;   // Read 32-bit long integer from reg[X], X = X + 1
  final public static int LREAD0    = 0x97;   // Read 32-bit long integer from reg[0]
  final public static int LREADBYTE = 0x98;   // Read lower 8 bits of reg[A]
  final public static int LREADWORD = 0x99;   // Read lower 16 bits reg[A]
  final public static int ATOL      = 0x9A;   // Convert ASCII to long integer
  final public static int LTOA      = 0x9B;   // Convert long integer to ASCII
  final public static int LSET      = 0x9C;   // reg[A] = reg[nn]
  final public static int LADD      = 0x9D;   // reg[A] = reg[A] + reg[nn]
  final public static int LSUB      = 0x9E;   // reg[A] = reg[A] - reg[nn]
  final public static int LMUL      = 0x9F;   // reg[A] = reg[A] * reg[nn]
  final public static int LDIV      = 0xA0;   // reg[A] = reg[A] / reg[nn]
  final public static int LCMP      = 0xA1;   // Signed long compare reg[A] - reg[nn]
  final public static int LUDIV     = 0xA2;   // reg[A] = reg[A] / reg[nn]
  final public static int LUCMP     = 0xA3;   // Unsigned long compare of reg[A] - reg[nn]
  final public static int LTST      = 0xA4;   // Long integer status of reg[A] AND reg[nn]
  final public static int LSET0     = 0xA5;   // reg[A] = reg[0]
  final public static int LADD0     = 0xA6;   // reg[A] = reg[A] + reg[0]
  final public static int LSUB0     = 0xA7;   // reg[A] = reg[A] - reg[0]
  final public static int LMUL0     = 0xA8;   // reg[A] = reg[A] * reg[0]
  final public static int LDIV0     = 0xA9;   // reg[A] = reg[A] / reg[0]
  final public static int LCMP0     = 0xAA;   // Signed long compare reg[A] - reg[0]
  final public static int LUDIV0    = 0xAB;   // reg[A] = reg[A] / reg[0]
  final public static int LUCMP0    = 0xAC;   // Unsigned long compare reg[A] - reg[0]
  final public static int LTST0     = 0xAD;   // Long integer status of reg[A] AND reg[0]
  final public static int LSETI     = 0xAE;   // reg[A] = long(bb)
  final public static int LADDI     = 0xAF;   // reg[A] = reg[A] + long(bb)
  final public static int LSUBI     = 0xB0;   // reg[A] = reg[A] - long(bb)
  final public static int LMULI     = 0xB1;   // reg[A] = reg[A] * long(bb)
  final public static int LDIVI     = 0xB2;   // reg[A] = reg[A] / long(bb)
  final public static int LCMPI     = 0xB3;   // Signed long compare reg[A] - long(bb)
  final public static int LUDIVI    = 0xB4;   // reg[A] = reg[A] / unsigned long(bb)
  final public static int LUCMPI    = 0xB5;   // Unsigned long compare reg[A] - ulong(bb)
  final public static int LTSTI     = 0xB6;   // Long integer status of reg[A] AND ulong(bb)
  final public static int LSTATUS   = 0xB7;   // Long integer status of reg[nn]
  final public static int LSTATUSA  = 0xB8;   // Long integer status of reg[A]
  final public static int LCMP2     = 0xB9;   // Signed long compare reg[nn] - reg[mm]
  final public static int LUCMP2    = 0xBA;   // Unsigned long compare reg[nn] - reg[mm]
  final public static int LNEG      = 0xBB;   // reg[A] = -reg[A]
  final public static int LABS      = 0xBC;   // reg[A] = | reg[A] |
  final public static int LINC      = 0xBD;   // reg[nn] = reg[nn] + 1
  final public static int LDEC      = 0xBE;   // reg[nn] = reg[nn] - 1
  final public static int LNOT      = 0xBF;   // reg[A] = NOT reg[A]
  final public static int LAND      = 0xC0;   // reg[A] = reg[A] AND reg[nn]
  final public static int LOR       = 0xC1;   // reg[A] = reg[A] OR reg[nn]
  final public static int LXOR      = 0xC2;   // reg[A] = reg[A] XOR reg[nn]
  final public static int LSHIFT    = 0xC3;   // reg[A] = reg[A] shift reg[nn]
  final public static int LMIN      = 0xC4;   // reg[A] = min(reg[A], reg[nn])
  final public static int LMAX      = 0xC5;   // reg[A] = max(reg[A], reg[nn])
  final public static int LONGBYTE  = 0xC6;   // reg[0] = long(signed byte bb)
  final public static int LONGUBYTE = 0xC7;   // reg[0] = long(unsigned byte bb)
  final public static int LONGWORD  = 0xC8;   // reg[0] = long(signed word wwww)
  final public static int LONGUWORD = 0xC9;   // reg[0] = long(unsigned word wwww)
  final public static int SETSTATUS = 0xCD;   // Set status byte
  final public static int SEROUT    = 0xCE;   // Serial output
  final public static int SERIN     = 0xCF;   // Serial Input
  final public static int SETOUT    = 0xD0;   // Set OUT1 and OUT2 output pins
  final public static int ADCMODE   = 0xD1;   // Set A/D trigger mode
  final public static int ADCTRIG   = 0xD2;   // A/D manual trigger
  final public static int ADCSCALE  = 0xD3;   // ADCscale[ch] = B
  final public static int ADCLONG   = 0xD4;   // reg[0] = ADCvalue[ch]
  final public static int ADCLOAD   = 0xD5;   // reg[0] = float(ADCvalue[ch]) * ADCscale[ch]
  final public static int ADCWAIT   = 0xD6;   // wait for next A/D sample
  final public static int TIMESET   = 0xD7;   // time = reg[0]
  final public static int TIMELONG  = 0xD8;   // reg[0] = time (long)
  final public static int TICKLONG  = 0xD9;   // reg[0] = ticks (long)
  final public static int EESAVE    = 0xDA;   // EEPROM[nn] = reg[mm]
  final public static int EESAVEA   = 0xDB;   // EEPROM[nn] = reg[A]
  final public static int EELOAD    = 0xDC;   // reg[nn] = EEPROM[mm]
  final public static int EELOADA   = 0xDD;   // reg[A] = EEPROM[nn]
  final public static int EEWRITE   = 0xDE;   // Store bytes in EEPROM
  final public static int EXTSET    = 0xE0;   // external input count = reg[0]
  final public static int EXTLONG   = 0xE1;   // reg[0] = external input counter (long)
  final public static int EXTWAIT   = 0xE2;   // wait for next external input
  final public static int STRSET    = 0xE3;   // Copy string to string buffer
  final public static int STRSEL    = 0xE4;   // Set selection point
  final public static int STRINS    = 0xE5;   // Insert string at selection point
  final public static int STRCMP    = 0xE6;   // Compare string with string buffer
  final public static int STRFIND   = 0xE7;   // Find string and set selection point
  final public static int STRFCHR   = 0xE8;   // Set field separators
  final public static int STRFIELD  = 0xE9;   // Find field and set selection point
  final public static int STRTOF    = 0xEA;   // Convert string selection to float
  final public static int STRTOL    = 0xEB;   // Convert string selection to long
  final public static int READSEL   = 0xEC;   // Read string selection
  final public static int STRBYTE   = 0xED;   // Insert 8-bit byte at selection point
  final public static int STRINC    = 0xEE;   // increment selection point
  final public static int STRDEC    = 0xEF;   // decrement selection point
  final public static int SYNC      = 0xF0;   // Get synchronization byte
  final public static int READSTATUS = 0xF1;  // Read status byte
  final public static int READSTR   = 0xF2;   // Read string from string buffer
  final public static int VERSION   = 0xF3;   // Copy version string to string buffer
  final public static int IEEEMODE  = 0xF4;   // Set IEEE mode (default)
  final public static int PICMODE   = 0xF5;   // Set PIC mode
  final public static int CHECKSUM  = 0xF6;   // Calculate checksum for uM-FPU code
  final public static int BREAK     = 0xF7;   // Debug breakpoint
  final public static int TRACEOFF  = 0xF8;   // Turn debug trace off
  final public static int TRACEON   = 0xF9;   // Turn debug trace on
  final public static int TRACESTR  = 0xFA;   // Send string to debug trace buffer
  final public static int TRACEREG  = 0xFB;   // Send register value to trace buffer
  final public static int READVAR   = 0xFC;   // Read internal variable, store in reg[0]
  final public static int RESET     = 0xFF;   // No operation (9 consecutive FF bytes cause a reset)

  final public static int SYNC_CHAR = 0x5C;   // sync character

//-------------------- uM-FPU V3.1 MOP definitions -----------------------------


  final public static int SCALAR_SET   = 0;   // MA[row,col] = reg[0], for all elements
  final public static int SCALAR_ADD   = 1;   // MA[row,col] = MA[row,col] + reg[0], for all elements
  final public static int SCALAR_SUB   = 2;   // MA[row,col] = reg[0] - MA[row,col], for all elements
  final public static int SCALAR_SUBR  = 3;   // MA[row,col] = MA[row,col] + reg[0], for all elements
  final public static int SCALAR_MUL   = 4;   // MA[row,col] = MA[row,col] * reg[0], for all elements
  final public static int SCALAR_DIV   = 5;   // MA[row,col] = MA[row,col] / reg[0], for all elements
  final public static int SCALAR_DIVR  = 6;   // MA[row,col] = reg[0] / MA[row,col], for all elements
  final public static int SCALAR_POW   = 7;   // MA[row,col] = MA[row,col] ** reg[0], for all elements
  final public static int EWISE_SET    = 8;   // MA[row,col] = MB[row,col], for all elements
  final public static int EWISE_ADD    = 9;   // MA[row,col] = MA[row,col] + MB[row,col], for all elements
  final public static int EWISE_SUB    = 10;  // MA[row,col] = MB[row,col] - MA[row,col], for all elements
  final public static int EWISE_SUBR   = 11;  // MA[row,col] = MA[row,col] + MB[row,col], for all elements
  final public static int EWISE_MUL    = 12;  // MA[row,col] = MA[row,col] * MB[row,col], for all elements
  final public static int EWISE_DIV    = 13;  // MA[row,col] = MA[row,col] / MB[row,col], for all elements
  final public static int EWISE_DIVR   = 14;  // MA[row,col] = MB[row,col] / MA[row,col], for all elements
  final public static int EWISE_POW    = 15;  // MA[row,col] = MA[row,col] ** MB[row,col], for all elements
  final public static int MX_MULTIPLY  = 16;  // MA = MB * MC matrix multiply
  final public static int MX_IDENTITY  = 17;  // MA = identity matrix
  final public static int MX_DIAGONAL  = 18;  // MA = diagonal matrix
  final public static int MX_TRANSPOSE = 19;  // MA = transpose of MB
  final public static int MX_COUNT     = 20;  // reg[0] = number of elements in MA
  final public static int MX_SUM       = 21;  // reg[0] = sum of all elements in MA
  final public static int MX_AVE       = 22;  // reg[0] = average value of all elements in MA
  final public static int MX_MIN       = 23;  // reg[0] = minimum value of all elements in MA
  final public static int MX_MAX       = 24;  // reg[0] = maximum value of all elements in MA
  final public static int MX_COPYAB    = 25;  // copy MA to MC
  final public static int MX_COPYAC    = 26;  // copy MA to MC
  final public static int MX_COPYBA    = 27;  // copy MB to MA
  final public static int MX_COPYBC    = 28;  // copy MB to MC
  final public static int MX_COPYCA    = 29;  // copy MC to MA
  final public static int MX_COPYCB    = 30;  // copy MC to MB
  final public static int MX_DETERM    = 31;  // reg0 = determinant of MA (2x2 and 3x3)
  final public static int MX_INVERSE   = 32;  // MA = inverse of MB (2x2 and 3x3)
  final public static int MX_ILOADRA   = 33;  // indexed load registers to MA
  final public static int MX_ILOADRB   = 34;  // indexed load registers to MB
  final public static int MX_ILOADRC   = 35;  // indexed load registers to MC
  final public static int MX_ILOADBA   = 36;  // indexed load MB to MA
  final public static int MX_ILOADCA   = 37;  // indexed load MC to MA
  final public static int MX_ISAVEAR   = 38;  // indexed save MA to registers
  final public static int MX_ISAVEAB   = 39;  // indexed save MA to MB
  final public static int MX_ISAVEAC   = 40;  // indexed save MA to MC

//-------------------- uM-FPU V3.1 FFT definitions -----------------------------

  final public static int FIRST_STAGE  = 0;   // first stage of FFT
  final public static int NEXT_STAGE   = 1;   // next stage of FFT
  final public static int NEXT_LEVEL   = 2;   // next level of FFT
  final public static int NEXT_BLOCK   = 3;   // next block of FFT
  final public static int BIT_REVERSE  = 4;   // bit reverse sort (pre-processing)
  final public static int PRE_ADJUST   = 8;   // inverse FFT (pre-processing)
  final public static int POST_ADJUST  = 16;  // inverse FFT (post-processing)

  // status flag definitions
  final public static int ZERO_FLAG     = 0x0001;   // zero flag
  final public static int SIGN_FLAG     = 0x0002;   // sign flag
  final public static int NAN_FLAG      = 0x0004;   // NaN flag
  final public static int INFINITY_FLAG = 0x0008;   // infinity flag

  // I/O pins for uM-FPU chip
  final static int CLOCK_PIN = CPU.pin15;       // default clock pin
  final static int DATA_PIN = CPU.pin14;        // default data pin
  protected static int fpuClockPin = 0;         // uM-FPU clock pin
  protected static int fpuDataPin = 0;          // uM-FPU data pin

  private static Float32 tmpFloat = new Float32(); // temporary Float32
  private static Int32 tmpLong  = new Int32();  // temporary Int32
  private static StringBuffer sbuf;             // string buffer (allocated as required)

  //-------------------- reset ------------------------------------------------
  /**
   * Reset the uM-FPU floating point coprocessor using specified data and
   * clock pins.
   * @param dataPin  SPI data pin
   * @param clockPin SPI clock pin
   * @return <code>true</code> if uM-FPU is detected and communication is confirmed.
   * <br><code>false</code> if uM-FPU is not detected or is not communicating.
   */
  public static boolean reset(int dataPin, int clockPin) {
    fpuDataPin = dataPin;
    fpuClockPin = clockPin;
    return(reset());
  }

  /**
   * Reset the uM-FPU floating point coprocessor.
   * @return <code>true</code> if uM-FPU is detected and communication is confirmed.
   * <br><code>false</code> if uM-FPU is not detected or is not communicating.
   */
  public static boolean reset() {
    // check for default reset
    if (fpuClockPin == 0) {
      fpuClockPin = CLOCK_PIN;
      fpuDataPin = DATA_PIN;
    }

    // send 10 reset bytes
    CPU.writePin(fpuClockPin, false);
    CPU.writePin(fpuDataPin, false);
    for (int i = 0; i < 10; i++) write(RESET);
    CPU.writePin(fpuDataPin, false);

    // delay for 10 msec
    CPU.delay(105);

    // check if FPU is responding
    if (CPU.readPin(fpuDataPin)) return(false);
    write(SYNC);
    int status = read();
    if (status == SYNC_CHAR)
      return(true);
    else
      return(false);
  }

  //-------------------- startWrite -------------------------------------------
  /**
   * Start a write transfer to the uM-FPU.
   */
  public static void startWrite() {
  }

  //-------------------- startRead --------------------------------------------
  /**
   * Start a read transfer from the uM-FPU.
   */
  public static void startRead() {
  }

    //-------------------- stop -----------------------------------------------
  /**
   * Stop a write or read transfer from the the uM-FPU.
   */
  public static void stop() {
  }

  //-------------------- write ------------------------------------------------
  /**
   * Write from 1 to 8 bytes to the uM-FPU. A method is provided for
   * each case, so only as many dataBytes as are needed need to be used.
   * e.g. write(1,2), or write(1,2,3,4), etc.
   * @param b1 8-bit data value.
   * @param b2 8-bit data value.
   * @param b3 8-bit data value.
   * @param b4 8-bit data value.
   * @param b5 8-bit data value.
   * @param b6 8-bit data value.
   * @param b7 8-bit data value.
   * @param b8 8-bit data value.
   */
  public static void write(int b1, int b2, int b3, int b4, int b5, int b6, int b7, int b8) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b3 << 8) + b4);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b5 << 8) + b6);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b7 << 8) + b8);
  }
  public static void write(int b1, int b2, int b3, int b4, int b5, int b6, int b7) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b3 << 8) + b4);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b5 << 8) + b6);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 8, CPU.SHIFT_MSB, (b7 << 8));
  }
  public static void write(int b1, int b2, int b3, int b4, int b5, int b6) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b3 << 8) + b4);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b5 << 8) + b6);
  }
  public static void write(int b1, int b2, int b3, int b4, int b5) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b3 << 8) + b4);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 8, CPU.SHIFT_MSB, (b5 << 8));
  }
  public static void write(int b1, int b2, int b3, int b4) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b3 << 8) + b4);
  }
  public static void write(int b1, int b2, int b3) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
    CPU.shiftOut(fpuDataPin, fpuClockPin, 8, CPU.SHIFT_MSB, (b3 << 8));
  }
  public static void write(int b1, int b2) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, (b1 << 8) + b2);
  }
  public static void write(int b1) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 8, CPU.SHIFT_MSB, (b1 << 8));
  }

  //-------------------- writeWord --------------------------------------------
  /**
   * Write 16-bit value to the uM-FPU.
   * @param dataWord 16-bit data value.
   */
  public static void writeWord(int dataWord) {
    CPU.shiftOut(fpuDataPin, fpuClockPin, 16, CPU.SHIFT_MSB, dataWord);
  }

  //-------------------- writeFloat -------------------------------------------
  /**
   * Write Float32 value to the uM-FPU.
   * @param fnum Float32 object.
   */
  public static void writeFloatA(Float32 fnum) {
    write(FWRITEA);
    writeFloat(fnum);
  }

  public static void writeFloatX(Float32 fnum) {
    write(FWRITEX);
    writeFloat(fnum);
  }

  public static void writeFloat(int reg, Float32 fnum) {
    if (reg == 0)
      write(FWRITE0);
    else
      write(FWRITE, reg);
      writeFloat(fnum);
  }

  private static void writeFloat(Float32 fnum) {
    writeWord(fnum.high);
    writeWord(fnum.low);
  }

  //-------------------- writeString ------------------------------------------
  /**
   * Write string to the uM-FPU.
   * @param dataWord 16-bit data value.
   */
  public static void writeString(String s) {
    // convert string to character array
    char sa[] = s.toCharArray();
    int len = s.length();
    int i = 0;

    // skip whitespace
    while (i < len && sa[i] == ' ' || sa[i] == '\t') i++;

    // send string
    while (i < len)
      CPU.shiftOut(fpuDataPin, fpuClockPin, 8, CPU.SHIFT_MSB, sa[i++] << 8);

    // send zero terminator
    CPU.shiftOut(fpuDataPin, fpuClockPin, 8, CPU.SHIFT_MSB, 0);
  }

  //-------------------- read -------------------------------------------------
  /**
   * Read an 8-bit value from the uM-FPU.
   * @return 8-bit data value.
   */
  public static int read() {
    return CPU.shiftIn(fpuDataPin, fpuClockPin, 8, CPU.PRE_CLOCK_MSB);
  }

  //-------------------- readWord ---------------------------------------------
  /**
   * Read a 16-bit value from the uM-FPU.
   * @return 16-bit data value.
   */
  public static int readWord() {
    return CPU.shiftIn(fpuDataPin, fpuClockPin, 16, CPU.PRE_CLOCK_MSB);
  }

  //-------------------- readFloat --------------------------------------------
  /**
   * Read 32-bit floating point value from uM-FPU register A.
   * @return Float32 32-bit floating point value.
   */
  public static Float32 readFloatA() {
    return(readFloat(FREADA, 0));
  }
  /**
   * Read 32-bit floating point value from uM-FPU register X.
   * @return Float32 32-bit floating point value.
   */
  public static Float32 readFloatX() {
    return(readFloat(FREADX, 0));
  }
  /**
   * Read 32-bit floating point value from the uM-FPU register.
   * @param reg register value.
   * @return Float32 32-bit floating point value.
   */
  public static Float32 readFloat(int reg) {
    int opcode;
    if (reg == 0)
      opcode = FREAD0;
    else
      opcode = FREAD;
    return(readFloat(opcode, reg));
  }
  private static Float32 readFloat(int opcode, int reg) {
    // wait until uM-FPU is ready, send command and read data
    wait();
    write(opcode);
    if (reg != 0) write(reg);
    tmpFloat.high = readWord();
    tmpFloat.low = readWord();
    return(tmpFloat);
  }

  //-------------------- readLong --------------------------------------------
  /**
   * Read 32-bit long integer value from uM-FPU register A.
   * @return Float32 32-bit floating point value.
   */
  public static Int32 readLongA() {
    return(readLong(LREADA, 0));
  }
  /**
   * Read 32-bit long integer value from uM-FPU register X.
   * @return Float32 32-bit floating point value.
   */
  public static Int32 readLongX() {
    return(readLong(LREADX, 0));
  }
  /**
   * Read 32-bit long integer value from the uM-FPU register.
   * @param reg register value.
   * @return Float32 32-bit floating point value.
   */
  public static Int32 readLong(int reg) {
    int opcode;
    if (reg == 0)
      opcode = FREAD0;
    else
      opcode = FREAD;
    return(readLong(opcode, reg));
  }
  private static Int32 readLong(int opcode, int reg) {
    // wait until uM-FPU is ready, send command and read data
    wait();
    write(opcode);
    if (reg != 0) write(reg);
    tmpLong.high = readWord();
    tmpLong.low = readWord();
    return(tmpLong);
  }

  //-------------------- readString -------------------------------------------
  /**
   * Read a string from the uM-FPU.
   * @return string read from uM-FPU.
   */

   public static String readString() {
    // allocate string buffer (if required)
    if (sbuf == null)
      sbuf = new StringBuffer(16);
    else
      sbuf.clear();

    // wait until uM-FPU is ready and send read string command
    wait();
    write(READSTR);

    // read the string and store in string buffer
    startRead();
    char ch;
    while (true) {
      ch = (char) read();
      if (ch == 0 || ch > 127) break;
      sbuf.append(ch);
    }
    return sbuf.toString();
  }

  //-------------------- wait -------------------------------------------------
  /**
   * Wait until uM-FPU is ready.
   */
  public static void wait() {
    // wait for the Fpu to be ready
    while(CPU.readPin(fpuDataPin)) ;
  }

  //-------------------- version ----------------------------------------------
  /**
   * Return the version string from the uM-FPU.
   * @return version string.
    */
  public static String version() {

    // return the version string
    startWrite();
    write(VERSION);
    return readString();
  }

  //-------------------- floatFormat ------------------------------------------
  /**
   * Select a uM-FPU register as the A register, and return the floating point
   * value in the register as a formatted string.
   * @param reg uM-FPU register.
   * @param format format specification.
   * @return string.
   */
  public static String floatFormat(int reg, int format) {
    startWrite();
    write(SELECTA, reg, FTOA, format);
    return readString();
  }

 /**
   * Return the floating point value of the currently selected uM-FPU register
   * (the A register), as a formatted string.
   * @param format format specification.
   * @return string.
   */
  public static String floatFormat(int format) {
    startWrite();
    write(FTOA, format);
    return readString();
  }

  /**
   * Return the floating point value of the currently selected uM-FPU register
   * (the A register), as an unformatted string.
   * @return string.
   */
  public static String floatFormat() {
    return floatFormat(0);
  }

  //-------------------- longFormat -------------------------------------------
  /**
   * Select a uM-FPU register as the A register, and return the long integer
   * value in the register as a formatted string.
   * @param reg uM-FPU register.
   * @param format format specification.
   * @return string.
   */
  public static String longFormat(int reg, int format) {
    startWrite();
    write(SELECTA, reg, LTOA, format);
    return readString();
  }

 /**
   * Return the long integer value of the currently selected uM-FPU register
   * (the A register), as a formatted string.
   * @param format format specification.
   * @return string.
   */
  public static String longFormat(int format) {
    startWrite();
    write(LTOA, format);
    return readString();
  }

  /**
   * Return the long integer value of the currently selected uM-FPU register
   * (the A register), as an unformatted string.
   * @return string.
   */
  public static String longFormat() {
    return longFormat(0);
  }

} // end class