/*
 * Copyright  Micromega Corporation 2002-2007. All rights reserved.
 */
package com.micromegacorp.math.v3_spi;
/**
 * @file    Math32.java
 * @target  Javelin Stamp
 *
 * This class provides math routines using Float32 numbers.
 *
 * @author Cam Thompson, Micromega Corporation
 * @version
 *    March 20, 2006
 *    - original version for uM-FPU V3
 */

 public class Math32 {

  private static Float32 result = new Float32();  // temporary Float32 result

  /**
   * Add two <code>Float32</code> values. Returns a + b.
   * @param a <code>Float32</code> value.
   * @param b <code>Float32</code> value.
   * @return <code>Float32</code> reference to result.
   */
  public static Float32 add(Float32 a, Float32 b) {
    result.set(a);
    return(result.add(b));
  }

  /**
   * Subtracts two <code>Float32</code> values. Returns a - b.
   * @param a <code>Float32</code> value.
   * @param b <code>Float32</code> value.
   * @return <code>Float32</code> reference to result.
   */
  public static Float32 subtract(Float32 a, Float32 b) {
    result.set(a);
    return(result.subtract(b));
  }

  /**
   * multiplies two <code>Float32</code> values. Returns a * b.
   * @param a <code>Float32</code> value.
   * @param b <code>Float32</code> value.
   * @return <code>Float32</code> reference to result.
   */
  public static Float32 multiply(Float32 a, Float32 b) {
    result.set(a);
    return(result.multiply(b));
  }

  /**
   * Divides a <code>Float32</code> value by another. Returns a / b.
   * @param a <code>Float32</code> value.
   * @param b <code>Float32</code> value.
   * @return <code>Float32</code> reference to result.
   */
  public static Float32 divide(Float32 a, Float32 b) {
    result.set(a);
    return(result.divide(b));
  }

    //-------------------- sin --------------------------------------------------
  /**
   * Returns the trigonometric sine of an angle.
   * Special cases:
   * <ul><li>If the argument is NaN or an infinity, then the result is NaN.
   * <li>If the argument is zero, then the result is a zero with the same
   * sign as the argument.
   * </ul>
   * @param angle angle in radians
   * @return <code>Float32</code> reference to sine of the angle
   */
  public static Float32 sin(Float32 angle) {
    return(command1(Fpu.SIN, angle));
  }

  //-------------------- cos --------------------------------------------------
  /**
   * Returns the trigonometric cosine of an angle.
   * Special cases:
   * <ul><li>If the argument is NaN or an infinity, then the result is NaN.
   * </ul>
   * @param angle angle in radians
   * @return <code>Float32</code> reference to cosine of the angle
   */
  public static Float32 cos(Float32 angle) {
    return(command1(Fpu.COS, angle));
  }

  //-------------------- tan --------------------------------------------------
  /**
   * Returns the trigonometric tangent of an angle.
   * Special cases:
   * <ul><li>If the argument is NaN or an infinity, then the result is NaN.
   * <li>If the argument is zero, then the result is a zero with the same
   * sign as the argument.
   * </ul>
   * @param angle angle in radians
   * @return <code>Float32</code> reference to tangent of the angle
   */
  public static Float32 tan(Float32 angle) {
    return(command1(Fpu.TAN, angle));
  }

  protected static Float32 command1(int opcode, Float32 fnum) {
    Fpu.startWrite();
    Fpu.write(Fpu.LEFT, Fpu.FWRITEA);
    Fpu.writeWord(fnum.high);
    Fpu.writeWord(fnum.low);
    Fpu.write(opcode, Fpu.RIGHT);
    Fpu.wait();
    Fpu.write(Fpu.FREAD, 0);
    result.high = Fpu.readWord();
    result.low = Fpu.readWord();
    return(result);
  }

}